/* Goalkeeper-Klasse ***********************************************
 *
 * Basisverhalten des Torwarts in "vW_Goalkeeper"
 * zustzliche Implementierungen in Rost, Brasas und
 * Borel
 * alte Version neu ueberarbeitet von Esko ab 31.7.2000
 *
 *******************************************************************
 */

#include "vW_torhueter.h"


/* --------------------------------------------------------------
 * Torhueter positioniert sich zwischen vermutetem Ball und Tor
 * --
 * Je nachdem, ob es sich um einen Dribbelangriff oder direkten
 * Torschuss handelt, positioniert sich der Torwart
 * --
 * falls es sich um einen Dribbelangriff handelt wird je nach
 * Einfallswinkel des Balls auf verschiedene Strategien zurueck-
 * gegriffen (Zentral,Links/Rechts oder Steil L/R)
 * --------------------------------------------------------------
 */


void vW_Goalkeeper::stand_between_predicted_Ball_and_Goal()
{
  Vector mypos = Mem->MyPos();                        // my actual position
  Vector blpos = Mem->BallAbsolutePosition();         // pos of ball right now
  Vector pdpos = Mem->BallPredictedPosition(2);       // pos of ball in 2 cycles
  Vector pppos = Mem->BallPredictedPosition(5);       // pos of ball in 5 cycles
  Vector gopos = Mem->MarkerPosition(Mem->RM_My_Goal);// center of the goal
  

  float coney = fabs(gopos.dist(blpos));             // dist between ball and goal center
  float gldis = fabs(MyGoalLeft.dist(blpos));        // dist between left post and ball
  float grdis = fabs(MyGoalRight.dist(blpos));       // dist between right post and goal
  float bally = mypos.dist(blpos);                   // dist between me and ball
  float goaldir = fabs(gopos.dist(pdpos)) - coney;   // negative-> ball is heading towards goal


  float  zone_c    = GOAL_CENTER_RANGE;                           // mid-center zone
  float  zone_m    = GOAL_MID_RANGE;                              // side zone
  bool   goalshot  = false;                                       // ball heading towards own goal
  float ang_goal   = calc_angle_relative_to(blpos,gopos);         // angle between goal center and ball 
  Line catch_line  = LineFromTwoPoints(mypos,mypos+Vector(0,10)); // Abfanglinie des Torwarts
  coney            = Min(coney,Min(gldis,grdis));       // real attack dist to goal(posts and center)
  
  if (coney < GOAL_DEFEND_ZONE) {   // falls der Ball in den Bereich des Torwarts kommt...
    float neg_dis = mypos.x-blpos.x;
    if (neg_dis>0.2f){   ///////////////////////////////////////////////////////////////////
      //cout << "BALL BEHIND ME" << endl;
      float v_cm = 0.5f;
      if (neg_dis*2>v_cm) { 
	v_cm = neg_dis/2; 
      }
      Vector point = Vector((-Mem->SP_pitch_length/2)+v_cm,pdpos.y);
      Line last_line = LineFromTwoPoints(gopos+Vector(v_cm,0),gopos+Vector(v_cm,10));
      Line shot_line = LineFromTwoPoints(pdpos,pppos);
      if (last_line.SameSlope(shot_line)){
	/// nothin
      } else {
	point = last_line.intersection(shot_line);
      }
      if (fabs(point.y)<(Mem->SP_goal_width/2+GOAL_NEARLY_HIT)){
	go_to_point(point,0,100,DT_none);
      } else {
	go_to_point(MyFavoritePoint,0,100,DT_none);
      }
    } else  {  //////////////////////////////////// not behind
      
    Unum unum  = Mem->ClosestOpponentTo(blpos);  // unum is closest opp to ball 
    if (unum != Unum_Unknown){                   // if there is any opp near to ball...  
      Vector opppos = Mem->OpponentAbsolutePosition(unum);    // opppos is pos of opp
      goalshot      = ((goaldir <0) && (fabs(opppos.dist(blpos))>GOAL_KICK_DIST));   /// OLD
      
      // true if ball heading towards goal  and opp close to ball
    }
    if (goalshot) {  // falls geschossen wird... 
      Line goal_line = Mem->FieldRectangle.LeftEdge();   // Torauslinie
      Line shot_line = LineFromTwoPoints(pdpos,pppos);   // Schusslinie des Balls
      
      if (shot_line.SameSlope(goal_line)){                   
	// cout << "GOAL KICK! BALL RUNNING PARALLEL TO GOAL LINE" << endl;
      } else 
	{
	  Vector v_hit  = goal_line.intersection(shot_line);      // Trefferposition

	  if (fabs(v_hit.dist(gopos))<(Mem->SP_goal_width/2+GOAL_NEARLY_HIT)){  
	    // Falls der Ball das Tor treffen wuerde...
	    
	    if (bally>GOAL_KICK_REACHABLE){ // falls der Ball weiter weg ist..
	      float v_cm = bally/2;
	      catch_line = LineFromTwoPoints(mypos+Vector(v_cm,0),mypos+Vector(v_cm,10));
	      // Abfanglinie 
	    } else if (bally<GOAL_KICK_UNREACHABLE){ // falls der Ball schon vorbei ist...
	      float v_cm = 0.7f;
	      catch_line = LineFromTwoPoints(MyGoalLeft+Vector(v_cm,0),MyGoalRight+Vector(v_cm,0));
	    }
	    
	    if (catch_line.SameSlope(shot_line)){
	      // cout << "GOAL KICK ERROR! Ball and Goalie on parallel line..."<< endl;
	    } else {
	      Vector v_inter  = catch_line.intersection(shot_line);  // smart point to intercept
	      //float re = calc_angle_relative_to(blpos,gopos);        // test!!!!!!!!
	      //Line re2 = LineFromTwoPoints(blpos,gopos);             // test!!!!!!!!
	      // cout << "GOAL KICK! intercept at ("<< v_inter.x << ","<< v_inter.y << ") G"<<
	      //	goal_line.angle() << ", B" << shot_line.angle() << ", (" << re << "," << re2.angle() << ")" <<
	      //	endl;
	      go_to_point(v_inter,0,100,DT_none);  // intercept ball at smart interception point
	    }
	  }
	  else {
	    // cout << "GOAL KICK! BALL WILL MISS GOAL"<<endl;
	    if (ang_goal>=-zone_c && ang_goal<=zone_c){    // if goal attacked central
	      Vector point = PointInBetween(Vector((-Mem->SP_pitch_length/2),0),pdpos,coney/2); // origin coney!
	      go_to_point(point,0,100,DT_none);
	    } 
	    if (ang_goal>=-zone_m && ang_goal<-zone_c){    // if goal attacked from right
	      go_to_point(PointInBetween(pdpos,MyCornerRight,4),0,100,DT_none);
	    }
	    if (ang_goal>zone_c && ang_goal<=zone_m){      // if goal attacked from left
	      go_to_point(PointInBetween(pdpos,MyCornerLeft,4),0,100,DT_none);
	    }
	    if (ang_goal<-zone_m){                         // if goal attacked from right corner
	      if (mypos.dist(MyCornerRight)>0.5f){
		go_to_point(MyCornerRight,0,100,DT_none);
	      } else {
		face_neck_and_body_to_point(blpos);
	      }
	    }
	    if (ang_goal>zone_m){                          // if goal attacked from left corner
	      if (mypos.dist(MyCornerLeft)>0.5f){
		go_to_point(MyCornerLeft,0,100,DT_none);
	      } else {
		face_neck_and_body_to_point(blpos);
	      }
	    }
	  }
	}
      
    } else 
      {
	goalkickcounter = 0;
	if (ang_goal>=-zone_c && ang_goal<=zone_c){    // if goal attacked central
	  Vector point = PointInBetween(Vector((-Mem->SP_pitch_length/2),0),pdpos,coney);  // this is the original
	  go_to_point(point,0,100,DT_none);
	} 
	if (ang_goal>=-zone_m && ang_goal<-zone_c){    // if goal attacked from right
	  go_to_point(PointInBetween(pdpos,MyCornerRight,4),0,100,DT_none);
	}
	if (ang_goal>zone_c && ang_goal<=zone_m){      // if goal attacked from left
	  go_to_point(PointInBetween(pdpos,MyCornerLeft,4),0,100,DT_none);
	}

	if (ang_goal<-zone_m){                         // if goal attacked from right corner
	  if (mypos.dist(MyCornerRight)>0.5f){
	    go_to_point(MyCornerRight,0,100,DT_none);
	  } else {
	    face_neck_and_body_to_point(blpos);
	  }
	}
	if (ang_goal>zone_m){                          // if goal attacked from left corner
	  if (mypos.dist(MyCornerLeft)>0.5f){
	    go_to_point(MyCornerLeft,0,100,DT_none);
	  } else {
	    face_neck_and_body_to_point(blpos);
	  }
	}
//  	if (ang_goal<-zone_m){                         // if goal attacked from right corner
//  	  go_to_point(MyCornerRight,0,100,DT_none);
//  	}
//  	if (ang_goal>zone_m){                          // if goal attacked from left corner
//  	  go_to_point(MyCornerLeft,0,100,DT_none);
//  	}
      }
    
    } //////////////////////////


  } else if (MyFavoritePoint.dist(mypos)>2) {
    go_to_point(MyFavoritePoint,0,90,DT_none);
  } else {
    if ((Mem->CurrentTime % 4) == 0)
      face_neck_and_body_to_point(Vector(0,0));
  }  
}





/* --------------------------------------------------------------
 * berechnet den Winkel zwischen zwei Punkten
 * --
 * z.B. zwischen Tormitte und Ball
 * --------------------------------------------------------------
 */

float vW_Goalkeeper::calc_angle_relative_to(Vector v1,Vector v2)
{
  Line l = LineFromTwoPoints(v1,v2);
  return l.angle();
//    Vector calcv   = v2 - v1;
//    float aang     = 1000;
//    if (calcv.x != 0) aang = (atan(calcv.y/calcv.x)/(2*M_PI)*360);
//    return aang;
}



/* --------------------------------------------------------------
 * berechnet den Winkel zwischen einem Angreifer und Tormitte
 * --------------------------------------------------------------
 */

float vW_Goalkeeper::calc_attack_angle(Unum unum) 
{
  if (unum==Unum_Unknown){
    my_error("There is no such attacker in angle");
    return 0;
  } else {
    return calc_angle_relative_to(Mem->OpponentAbsolutePosition(unum),
				  Mem->MarkerPosition(Mem->RM_My_Goal));
  }
}


/* --------------------------------------------------------------
 * berechnet den Winkel zwischen Angreifer und Tormitte
 * --------------------------------------------------------------
 */


float vW_Goalkeeper::calc_intercept_angle(Unum unum)
{
  if (unum==Unum_Unknown){
    my_error("calc_intercept_angle(unum): There is no such attacker in angle");
    return 365;
  } else {
    return calc_intercept_angle(Mem->OpponentAbsolutePosition(unum));
  }
}


/* --------------------------------------------------------------
 * berechnet den Winkel zwischen Angreifer und Tormitte
 * --------------------------------------------------------------
 */

float vW_Goalkeeper::calc_intercept_angle(Vector v)
{
  float aang     = calc_angle_relative_to(v,Mem->MarkerPosition(Mem->RM_My_Goal));
  float clang    = calc_angle_relative_to(v,MyGoalLeft);
  float crang    = calc_angle_relative_to(v,MyGoalRight);
  float s_ang    = -1;
  if ((clang<0) && (crang<0)) {
    s_ang = 180 - (180+clang-crang);
  } else if ((clang>0) && (crang<0)){
    s_ang = 180 - (clang-crang);
  } else {
    s_ang = 180 - (clang+180-crang);
    s_ang /= 2;
  }
  // cout << "intercept angle a"<<aang<<" l"<<clang<<" r"<<crang<<" =>"<<s_ang<<endl;
  return s_ang;
}



/* --------------------------------------------------------------
 * bla
 * --------------------------------------------------------------
 */

float vW_Goalkeeper::calc_opt_intercept_angle(Vector v)
{
  float clang = calc_angle_relative_to(v,MyGoalLeft);  
  float nang  = calc_intercept_angle(v);
  float sang  = 365;
  if (nang>0){
    sang = (clang + nang);
  } else {
    sang = (nang + (180 + clang));
  }
  return sang;
}



/* --------------------------------------------------------------
 * bla
 * --------------------------------------------------------------
 */

Vector vW_Goalkeeper::calc_opt_intercept_position(Vector v)
{
  Vector rv = Vector(-Mem->SP_pitch_length/2,0);
  float sang  = calc_opt_intercept_angle(v);
  float bang  = sang / 360 * (2*M_PI);
  rv.x = -Mem->SP_pitch_length/2-v.x;
  rv.y = v.y+ (tan(bang)* (-Mem->SP_pitch_length/2-v.x));
  rv.x = -Mem->SP_pitch_length/2;
  // cout << " estimated angle interception: "<< sang << " / " << rv.x << " " <<  rv.y << endl;
  return rv;
}




/** -------------------------------------------------
 * Einen Abschlag vorbereiten,
 * zunaechst warten, anschliessend auf die Seite 
 * gehen, auf der mehr eigene Spieler stehen
 * --------------------------------------------------
 */


void vW_Goalkeeper::initiate_cool_kick_off()
{
  float PLAYER_NEARBY_RANGE = KICK_OFF_TEAM_DIST;  
  if (kickcounter<WAIT_KICK_OFF_1) {
    if ((kickcounter % 2) ==0) scan_field_with_body();
    //cout << "INIT scan_field_with_body" << kickcounter << endl;
    kickcounter++;
  } else {
    //    Unum ntctr = Mem->NumTeammatesWithin(PLAYER_NEARBY_RANGE,MyKickInRight);
    //    Unum ntctl = Mem->NumTeammatesWithin(PLAYER_NEARBY_RANGE,MyKickInLeft);
    Unum opr = Mem->NumOpponentsWithin(PLAYER_NEARBY_RANGE,MyKickInRight);
    Unum opl = Mem->NumOpponentsWithin(PLAYER_NEARBY_RANGE,MyKickInLeft);
    Unum tmr = Mem->NumTeammatesWithin(PLAYER_NEARBY_RANGE,MyKickInRight);
    Unum tml = Mem->NumTeammatesWithin(PLAYER_NEARBY_RANGE,MyKickInLeft);
    if (opl==0){                           // no opp on the left...
      move(MyKickInLeft);                  // ... go to the left
    } else
      if (opr==0){                         // no opp on the right... 
	move(MyKickInRight);               // ... go to the right
	// cout << "MOVED LEFT BECAUSE NO OPPONENTS"<<endl;
      } else {
	if ((tml>0) && (tmr>0)){           // if teammates nearby check ...
	  float fr = opr/tmr;
	  float fl = opl/tml;
	  // cout << "FR "<<fr<<" FL "<<fl<<endl;
	  if (fr < fl){                    // if more own then opp on right... 
	    // cout << "MOVED RIGHT BECAUSE LESS OPPONENTS THAN TEAMMATES: ("<<fr<<"/"<<fl<<")"<<endl;
	    move(MyKickInRight);           // ... go to right
	  } else {                         // if more own then opp on left...                    
	    // cout << "MOVED LEFT BECAUSE LESS OPPONENTS THAN TEAMMATES: ("<<fr<<"/"<<fl<<")"<<endl;
	    move(MyKickInLeft);            // ... go to the left
	  }
	} else if (tml>0){                 // else if more own on the left
	// cout << "MOVED LEFT BECAUSE MORE TEAMMATES"<<endl;
	  move(MyKickInLeft);              // ... go to the left
	} else {                           // else if more own on the right
	// cout << "MOVED RIGHT BECAUSE MORE TEAMMATES"<<endl;
	  move(MyKickInRight);             // ... go to the right
	}
      }
    perform_kick_off = true;        // set flag for goal kick execution
    after_kick_off   = false;       // reset go-home flag 
    kickcounter      = 0;           // reset kickcounter   
  }
}

/** -----------------------------------------------------
 * Einen Abschlag durchfuehren,
 * erst passen, wenn ein eigener Spieler anspielbar ist
 * ------------------------------------------------------
 */

void vW_Goalkeeper::execute_cool_kick_off()
{
  if (kickcounter<WAIT_KICK_OFF_2) {  // wait for some time and scan field
    if ((kickcounter % 2) ==0)
      scan_field_with_body();
    //cout << "EXEC -> scan_field_with_body" << kickcounter << endl;
    kickcounter++;
  } else {                            // search teammate and pass ball
    TeammateToPassBall = findTeammateToPassBall(150);
    if (TeammateToPassBall){
      pass_ball(TeammateToPassBall);
      // cout << "EXEC -> PASSED THE BALL!"<< kickcounter << endl;
      perform_kick_off = false;
      kickcounter      = 0;
      bakocounter      = 0;
      after_kick_off   = true;
    } else {                          // still search the field
      scan_field_with_body();
      //cout << "EXEC -> Still scanning"<< kickcounter << endl;
      kickcounter++;
    } 
  }
}


/** -----------------------------------------------------
 * Einen Abschlag abschliessen,
 * schnell zurueck zum Tor gehen
 * ------------------------------------------------------
 */

void vW_Goalkeeper::behave_after_kick_off()
{
  if (bakocounter < 2){                  // pass again after pass
    pass_ball(TeammateToPassBall);
    bakocounter++;
  } else
    if (bakocounter < WAIT_KICK_OFF_3){  // while not finished return home
      go_to_point(MyFavoritePoint,0,100,DT_none);
      bakocounter++;
    } else {                             // finish goal kick off
      bakocounter=0;                      
      after_kick_off = false;
    } 
  //cout << "EXEC -> BEHAVE "<< bakocounter <<endl;
}


/** -----------------------------------------------------
 * Verhalten beim Einwurf 
 * je nach Einwurfseite auf diese Seite wechseln
 * ------------------------------------------------------
 */

void vW_Goalkeeper::behave_on_kick_in()
{
  Vector blpos = Mem->BallAbsolutePosition();  // NEW NEW NEW
  if (Mem->BallAbsolutePosition().x<KICK_IN_NEARBY)   // if ball nearby then  // NEW NEW NEW
    {       
      if (Mem->BallAbsolutePosition().y>0){           // go to side where kick in be executed...
	if (MyCornerLeft.dist(Mem->MyPos())>1)
	  {  
	    go_to_point(MyCornerLeft);
	  } else {
	    //scan_field_with_neck();   // !!!!!!!!!!!!!!!!!! Verbessern -> auf den Ball gucken
	    face_neck_and_body_to_point(blpos);  // NEW NEW NEW
	  }
      } else {
	if (MyCornerRight.dist(Mem->MyPos())>1) 
	  {
	    go_to_point(MyCornerRight);
	  } else {
	    //scan_field_with_neck();   // !!!!!!!!!!!!!!!!!! Verbessern -> auf den Ball gucken
	    face_neck_and_body_to_point(blpos);  // NEW NEW NEW
	  }
      }
    } else {
      go_to_point(MyFavoritePoint);
    }
}

/** -----------------------------------------------------
 * Verhalten bei einem Torabschlag
 * ein bisschen rumzittern
 * ------------------------------------------------------
 */
void vW_Goalkeeper::behave_on_my_goal_kick()
{
  GOAL_KICK_MOVER %= 3000;   // remain under 3000 (no special reason )
  if ((GOAL_KICK_MOVER % 8)==0){
    go_to_point(MyFavoritePoint+Vector(0,0.5f));
  } else 
    if ((GOAL_KICK_MOVER % 8)==4){
      go_to_point(MyFavoritePoint+Vector(0,-0.5f));
    } 
    else {
      face_neck_and_body_to_point(Vector(0,0));
    }
  GOAL_KICK_MOVER++;
}

/** -----------------------------------------------------
 * Verhalten bei einer Ecke
 * je nach Ecke auf diese Seite wechseln
 * ------------------------------------------------------
 */


void vW_Goalkeeper::behave_on_their_corner_kick()
{
  Vector blpos = Mem->BallAbsolutePosition();  // NEW NEW NEW
  if (Mem->BallAbsolutePosition().y>0){
    if (MyCornerLeft.dist(Mem->MyPos())>1) 
      {
	go_to_point(MyCornerLeft);
      } else {
	//scan_field_with_neck();      // !!!!!!!!!!!!!!!!!! Verbessern -> auf den Ball gucken
	face_neck_and_body_to_point(blpos);  // NEW NEW NEW
      }
  } else {
    if (MyCornerRight.dist(Mem->MyPos())>1) 
      {
	go_to_point(MyCornerRight);
      } else {
	//scan_field_with_neck();      // !!!!!!!!!!!!!!!!!! Verbessern -> auf den Ball gucken
	face_neck_and_body_to_point(blpos);  // NEW NEW NEW
      }
  }
} 

/** -----------------------------------------------------
 * normales Spielverhalten 
 * 
 * ------------------------------------------------------
 */


void vW_Goalkeeper::behave_default()
{
  if(Mem->BallCatchable()){
    goalie_catch( Mem->BallAngleFromBody() );
  } else {
    if (!Mem->FieldRectangle.IsWithin(Mem->MyPos())){
      // cout << "GOALIE ERROR -> Goalie left field !" << endl;
      go_to_point(MyFavoritePoint);
    } else {
      stand_between_predicted_Ball_and_Goal();
    }
  }  
} 




/* ------------------------------------------------------
 * Implementierung des neuen VirtualWerder Goalies Brasas
 * 31.7.2000 Esko
 * ------------------------------------------------------
 */



void Brasas::Behave()  // Brasas benimm dich :-)
{
  if (Mem->BallPositionValid())   // falls der Ball sichtbar ist...
    {
      if (Mem->MyNumber == Mem->ClosestTeammateTo(Mem->BallAbsolutePosition())) say("");
      if (after_kick_off) {    
	behave_after_kick_off();
      } else
	if (perform_kick_off) {
	  execute_cool_kick_off();
	} else 
	  
	  switch(Mem->PlayMode){    // in case of play mode ... do
	    
	  case PM_Before_Kick_Off:  // Anstossverhalten
	    {
	      move(MyFavoritePoint);
	      return;
	      break;
	    }
	  case PM_My_Goalie_Free_Kick:  // Abstoss
	    {
	      initiate_cool_kick_off();
	      return;
	      break;
	    }
	  case PM_My_Goal_Kick:
	    {
	      behave_on_my_goal_kick();  // NEW NEW NEW NEW  
	      //kick_ball(Vector(0,(Mem->SP_pitch_width/2)),KM_HardestKick,TURN_NONE);
	      return;
	      break;
	    }
	  case PM_My_Free_Kick:
	    {
	      if (MyFavoritePoint.dist(Mem->MyPos())>1){   // NEW NEW NEW NEW  
		go_to_point(MyFavoritePoint);} else
		  face_neck_and_body_to_point(Vector(0,0));// NEW NEW NEW NEW  
	      //execute_cool_kick_off();
	      return;
	      break;
	    }
	  case PM_My_Kick_In:       // Einwurf
	    {
	      behave_on_kick_in();
	      break;
	    }
	  case PM_Their_Kick_In:    // gegnerischer Einwurf
	    {
	      behave_on_kick_in();
	      break;
	    }
	  case PM_My_Corner_Kick:   // unser Einwurf
	    {
	      if (MyFavoritePoint.dist(Mem->MyPos())>1){   // NEW NEW NEW NEW  
		go_to_point(MyFavoritePoint,20);} else
		  face_neck_and_body_to_point(Vector(0,0));// NEW NEW NEW NEW  
	      break;
	    } 
	  case PM_Their_Corner_Kick:
	    {
	      behave_on_their_corner_kick();
	      break;	  
	    } 
	  default: // ansonsten
	    {
	      behave_default();
	    } 
	  } 
    }
  else
    scan_field_with_body();
}


void Borel::Behave()
{
  // nothing here yet
}

void vW_Goalkeeper::Behave(){

  if (Mem->BallPositionValid())
    {
      if (Mem->MyNumber == 
	  Mem->ClosestTeammateTo(Mem->BallAbsolutePosition()))
	{
	  say("");
	}
      
      
      switch(Mem->PlayMode){
	
      case PM_Before_Kick_Off:
	{
	  // Anstossverhalten
	  move(Mem->MarkerPosition(Mem->RM_My_Goal));
	  return;
	  break;
	}
	
      case PM_My_Goalie_Free_Kick:
	{
	  
	  if(fabs(Mem->MyPos().x-Mem->MyLeftGoalKickSpot.x)>1
	     && fabs(Mem->MyPos().y-Mem->MyLeftGoalKickSpot.y)>1){
	    
	    move(Mem->MyLeftGoalKickSpot);
	    
	  } else {
	    
	    
	    int TeammateToPassBall=findTeammateToPassBall();
		if(TeammateToPassBall){
		  
		  kick(PassToTeammateKickPower(TeammateToPassBall),
		       Mem->AngleToFromBody(Mem->TeammateAbsolutePosition(TeammateToPassBall)));
		  
		  return;
		  
		}
		
		kick(100,Mem->MarkerAngleFromBody(Mem->RM_LC_Flag));
	  }
	    break;
	}
	
	default:
	  {
	    if(Mem->BallCatchable()){
	      
		// Fass!
	      goalie_catch( Mem->BallAngleFromBody() );
	      
	    } else {
	      
	      stand_between_Ball_and_Goal();
	      
	    }
	  }
      }
    }
  else
    scan_field_with_body();
  
}


/* ------------------------------------------------
 * Torhueter positioniert sich zwischen Ball und Tor
 *
 */

void vW_Goalkeeper::stand_between_Ball_and_Goal()
{	
  // nothing in here		
}



/* ------------------------------------------------------------------
 * ???
 * ------------------------------------------------------------------
 */

Bool vW_Goalkeeper::isTeammateBetweenBallAndMe(){
  
  Line line;

  for (int i=0; i<Mem->SP_team_size; i++){  
    line = LineFromTwoPoints( Mem->BallAbsolutePosition(), Mem->MyPos() );
    if( i!=Mem->MyNumber )
      if( fabs( line.dist( Mem->TeammateAbsolutePosition(i) ) < 1) &&
	  Mem->MyPos().dist( Mem->BallAbsolutePosition() ) > 
	  Mem->MyPos().dist( Mem->TeammateAbsolutePosition(i) ) ){
	if(Mem->MyLocationSide()==FS_Right) 
	  my_error("TORWART Noch andere Mitspieler. %d",
		   line.dist( Mem->TeammateAbsolutePosition(i).x ));
	return TRUE;
      }
  }
  return FALSE;  
}


/*
 *
 *
 *
 */

void vW_Goalkeeper::gotoPredictedBallPosition()
{			
  Vector point=Mem->BallPredictedPosition(5);
  
  if(point.x < Mem->MarkerPosition(Mem->RM_My_Goal).x+3){
    point.x=Mem->MarkerPosition(Mem->RM_My_Goal).x+1;
  } else {
    if(point.x>=Mem->MarkerPosition(Mem->RM_My_Goal).x+16){
      point.x=10;
    }
  }
  if( point.y>7 )
    {
      point.y=7;
    } else {
      if (point.y<-7){
	point.y=-7;
      }
    }
  go_to_point(point,1,100,DT_none);
}





